/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_GFX_CAMERA_HPP
#define FGGL_GFX_CAMERA_HPP

#include "fggl/math/types.hpp"
#include "fggl/entity/entity.hpp"

namespace fggl::gfx {

	struct Camera {
		constexpr const static char name[] = "Camera";
		math::vec3 target = math::vec3(0.0f, 0.0f, 0.0f);
		float aspectRatio = 1280.0f / 720.0f;
		float fov = glm::radians(45.0f);
		float nearPlane = 0.1f;
		float farPlane = 100.0f;
	};

	inline math::mat4 calc_proj_matrix(const Camera& camera) {
		return glm::perspective(camera.fov, camera.aspectRatio, camera.nearPlane, camera.farPlane);
	}

	inline math::Ray get_camera_ray(const entity::EntityManager& world, const entity::EntityID camera, math::vec2 position) {
		auto& camTransform = world.get<fggl::math::Transform>(camera);
		auto& camComp = world.get<fggl::gfx::Camera>(camera);

		const auto projMatrix = fggl::gfx::calc_proj_matrix(camComp);
		const auto viewMatrix = fggl::math::calc_view_matrix(camTransform);

		glm::vec4 startNDC {
			position.x,
			position.y,
			-1.0f,
			1.0f
		};
		glm::vec4 endNDC {
			position.x,
			position.y,
			0.0f,
			1.0f
		};

		fggl::math::mat4 M = glm::inverse( projMatrix * viewMatrix );
		glm::vec3 start = M * startNDC;
		glm::vec3 end = M * endNDC;
		return { start, glm::normalize(end - start) };
	}

};

#endif
